package javax.swing.text;

import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.UndoableEdit;
import java.util.Vector;
import java.io.Serializable;

public final class StringContent implements AbstractDocument.Content, Serializable
{
    private static final char[] empty;
    private char[] data;
    private int count;
    transient Vector<StringContent.PosRec> marks;
    
    public StringContent() {
        this(10);
    }
    
    public StringContent(int n) {
        if (n < 1) {
            n = 1;
        }
        (this.data = new char[n])[0] = '\n';
        this.count = 1;
    }
    
    @Override
    public int length() {
        return this.count;
    }
    
    @Override
    public UndoableEdit insertString(int n, String s) throws BadLocationException {
        if (n >= this.count || n < 0) {
            throw new BadLocationException("Invalid location", this.count);
        }
        char[] charArray = s.toCharArray();
        this.replace(n, 0, charArray, 0, charArray.length);
        if (this.marks != null) {
            this.updateMarksForInsert(n, s.length());
        }
        return new StringContent.InsertUndo(n, s.length());
    }
    
    @Override
    public UndoableEdit remove(int n, int n2) throws BadLocationException {
        if (n + n2 >= this.count) {
            throw new BadLocationException("Invalid range", this.count);
        }
        StringContent.RemoveUndo removeUndo = new StringContent.RemoveUndo(n, this.getString(n, n2));
        this.replace(n, n2, StringContent.empty, 0, 0);
        if (this.marks != null) {
            this.updateMarksForRemove(n, n2);
            return removeUndo;
        }
        return removeUndo;
    }
    
    @Override
    public String getString(int offset, int count) throws BadLocationException {
        if (offset + count > this.count) {
            throw new BadLocationException("Invalid range", this.count);
        }
        return new String(this.data, offset, count);
    }
    
    @Override
    public void getChars(int offset, int count, Segment segment) throws BadLocationException {
        if (offset + count > this.count) {
            throw new BadLocationException("Invalid location", this.count);
        }
        segment.array = this.data;
        segment.offset = offset;
        segment.count = count;
        return;
    }
    
    @Override
    public Position createPosition(int n) throws BadLocationException {
        if (this.marks == null) {
            this.marks = new Vector();
        }
        return new StringContent.StickyPosition(n);
    }
    
    void replace(int n, int n2, char[] array, int n3, int n4) {
        int n5 = n4 - n2;
        int n6 = n + n2;
        int n7 = this.count - n6;
        int n8 = n6 + n5;
        if (this.count + n5 >= this.data.length) {
            char[] data = new char[Math.max(2 * this.data.length, this.count + n5)];
            System.arraycopy(this.data, 0, data, 0, n);
            System.arraycopy(array, n3, data, n, n4);
            System.arraycopy(this.data, n6, data, n8, n7);
            this.data = data;
        }
        else {
            System.arraycopy(this.data, n6, this.data, n8, n7);
            System.arraycopy(array, n3, this.data, n, n4);
        }
        this.count += n5;
    }
    
    void resize(int a) {
        char[] data = new char[a];
        System.arraycopy(this.data, 0, data, 0, Math.min(a, this.count));
        this.data = data;
    }
    
    synchronized void updateMarksForInsert(int n, int n2) {
        if (n == 0) {
            n = 1;
        }
        for (int size = this.marks.size(), i = 0; i < size; ++i) {
            StringContent.PosRec posRec = (StringContent.PosRec)this.marks.elementAt(i);
            if (posRec.unused) {
                this.marks.removeElementAt(i);
                --i;
                --size;
            }
            else if (posRec.offset >= n) {
                posRec.offset += n2;
            }
        }
    }
    
    synchronized void updateMarksForRemove(int offset, int n) {
        for (int size = this.marks.size(), i = 0; i < size; ++i) {
            StringContent.PosRec posRec = (StringContent.PosRec)this.marks.elementAt(i);
            if (posRec.unused) {
                this.marks.removeElementAt(i);
                --i;
                --size;
            }
            else if (posRec.offset >= offset + n) {
                posRec.offset -= n;
            }
            else if (posRec.offset >= offset) {
                posRec.offset = offset;
            }
        }
    }
    
    protected Vector getPositionsInRange(Vector vector, int n, int n2) {
        int size = this.marks.size();
        int n3 = n + n2;
        vector = ((vector == null) ? new Vector() : vector);
        for (int i = 0; i < size; ++i) {
            StringContent.PosRec posRec = (StringContent.PosRec)this.marks.elementAt(i);
            if (posRec.unused) {
                this.marks.removeElementAt(i);
                --i;
                --size;
            }
            else if (posRec.offset >= n && posRec.offset <= n3) {
                vector.addElement(new StringContent.UndoPosRef(posRec));
            }
        }
        return vector;
    }
    
    protected void updateUndoPositions(Vector vector) {
        for (int i = vector.size() - 1; i >= 0; --i) {
            StringContent.UndoPosRef undoPosRef = (StringContent.UndoPosRef)vector.elementAt(i);
            if (undoPosRef.rec.unused) {
                vector.removeElementAt(i);
            }
            else {
                undoPosRef.resetLocation();
            }
        }
    }
    
    static {
        empty = new char[0];
    }
    
    final class PosRec
    {
        int offset;
        boolean unused;
        
        PosRec(int offset) {
            this.offset = offset;
        }
    }
    
    final class StickyPosition implements Position
    {
        StringContent.PosRec rec;
        
        StickyPosition(int n) {
            this.rec = new StringContent.PosRec(n);
            StringContent.this.marks.addElement(this.rec);
        }
        
        @Override
        public int getOffset() {
            return this.rec.offset;
        }
        
        @Override
        protected void finalize() throws Throwable {
            this.rec.unused = true;
        }
        
        @Override
        public String toString() {
            return Integer.toString(this.getOffset());
        }
    }
    
    final class UndoPosRef
    {
        protected int undoLocation;
        protected StringContent.PosRec rec;
        
        UndoPosRef(StringContent.PosRec rec) {
            this.rec = rec;
            this.undoLocation = rec.offset;
        }
        
        protected void resetLocation() {
            this.rec.offset = this.undoLocation;
        }
    }
    
    class InsertUndo extends AbstractUndoableEdit
    {
        protected int offset;
        protected int length;
        protected String string;
        protected Vector posRefs;
        
        protected InsertUndo(int offset, int length) {
            this.offset = offset;
            this.length = length;
        }
        
        @Override
        public void undo() throws CannotUndoException {
            super.undo();
            try {
                synchronized (StringContent.this) {
                    if (StringContent.this.marks != null) {
                        this.posRefs = StringContent.this.getPositionsInRange(null, this.offset, this.length);
                    }
                    this.string = StringContent.this.getString(this.offset, this.length);
                    StringContent.this.remove(this.offset, this.length);
                }
            }
            catch (BadLocationException ex) {
                throw new CannotUndoException();
            }
        }
        
        @Override
        public void redo() throws CannotRedoException {
            super.redo();
            try {
                synchronized (StringContent.this) {
                    StringContent.this.insertString(this.offset, this.string);
                    this.string = null;
                    if (this.posRefs != null) {
                        StringContent.this.updateUndoPositions(this.posRefs);
                        this.posRefs = null;
                    }
                }
            }
            catch (BadLocationException ex) {
                throw new CannotRedoException();
            }
        }
    }
    
    class RemoveUndo extends AbstractUndoableEdit
    {
        protected int offset;
        protected int length;
        protected String string;
        protected Vector posRefs;
        
        protected RemoveUndo(int offset, String string) {
            this.offset = offset;
            this.string = string;
            this.length = string.length();
            if (StringContent.this.marks != null) {
                this.posRefs = StringContent.this.getPositionsInRange(null, offset, this.length);
            }
        }
        
        @Override
        public void undo() throws CannotUndoException {
            super.undo();
            try {
                synchronized (StringContent.this) {
                    StringContent.this.insertString(this.offset, this.string);
                    if (this.posRefs != null) {
                        StringContent.this.updateUndoPositions(this.posRefs);
                        this.posRefs = null;
                    }
                    this.string = null;
                }
            }
            catch (BadLocationException ex) {
                throw new CannotUndoException();
            }
        }
        
        @Override
        public void redo() throws CannotRedoException {
            super.redo();
            try {
                synchronized (StringContent.this) {
                    this.string = StringContent.this.getString(this.offset, this.length);
                    if (StringContent.this.marks != null) {
                        this.posRefs = StringContent.this.getPositionsInRange(null, this.offset, this.length);
                    }
                    StringContent.this.remove(this.offset, this.length);
                }
            }
            catch (BadLocationException ex) {
                throw new CannotRedoException();
            }
        }
    }
}
